<?php

namespace Drupal\ultimenu;

use Drupal\Component\Utility\Html;
use Drupal\Core\Url;
use Drupal\Core\Template\Attribute;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Block\BlockManagerInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Render\RendererInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Manages Ultimenu plugin.
 *
 * @see plugin_api
 */
class UltimenuManager extends UltimenuBase implements UltimenuManagerInterface {

  /**
   * Module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * Static cache for the menu blocks.
   *
   * @var array
   */
  protected $menuBlocks;

  /**
   * Static cache for the blocks.
   *
   * @var array
   */
  protected $blocks;

  /**
   * Static cache for the regions.
   *
   * @var array
   */
  protected $regions;

  /**
   * Static cache for the enabled regions.
   *
   * @var array
   */
  protected $enabledRegions;

  /**
   * Static cache for the menu options.
   *
   * @var array
   */
  protected $menuOptions;

  /**
   * The Ultimenu tree service.
   *
   * @var \Drupal\ultimenu\UltimenuTree
   */
  protected $tree;

  /**
   * The Ultimenu tool service.
   *
   * @var \Drupal\ultimenu\UltimenuTool
   */
  protected $tool;

  /**
   * Constructs a Ultimenu object.
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager, BlockManagerInterface $block_manager, ModuleHandlerInterface $module_handler, RendererInterface $renderer, UltimenuTreeInterface $tree, UltimenuToolInterface $tool) {
    parent::__construct($config_factory, $entity_type_manager, $block_manager);
    $this->moduleHandler = $module_handler;
    $this->renderer = $renderer;
    $this->tree = $tree;
    $this->tool = $tool;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.block'),
      $container->get('module_handler'),
      $container->get('renderer'),
      $container->get('ultimenu.tree'),
      $container->get('ultimenu.tool')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getModuleHandler() {
    return $this->moduleHandler;
  }

  /**
   * Returns the renderer.
   */
  public function getRenderer() {
    return $this->renderer;
  }

  /**
   * Returns the tool service.
   */
  public function getTool() {
    return $this->tool;
  }

  /**
   * Returns the tool service.
   */
  public function getTree() {
    return $this->tree;
  }

  /**
   * {@inheritdoc}
   */
  public function getMenus() {
    if (!isset($this->menuOptions)) {
      $this->menuOptions = $this->tree->getMenus();
    }
    return $this->menuOptions;
  }

  /**
   * {@inheritdoc}
   */
  public function getUltimenuBlocks() {
    if (!isset($this->menuBlocks)) {
      $this->menuBlocks = [];
      foreach ($this->getMenus() as $delta => $name) {
        if ($this->getEnabledBlocks($delta)) {
          $this->menuBlocks[$delta] = $this->t('@name', ['@name' => $name]);
        }
      }
      asort($this->menuBlocks);
    }
    return $this->menuBlocks;
  }

  /**
   * {@inheritdoc}
   */
  public function getEnabledBlocks($delta) {
    $blocks = $this->getSetting('blocks');
    return !empty($blocks[$delta]);
  }

  /**
   * {@inheritdoc}
   */
  public function build(array $build = []) {
    $build = [
      '#theme'      => 'ultimenu',
      '#items'      => [],
      '#build'      => $build,
      '#pre_render' => [[$this, 'preRenderBuild']],
    ];

    $this->moduleHandler->alter('ultimenu_build', $build);
    return $build;
  }

  /**
   * Builds the Ultimenu outputs as a structured array ready for ::renderer().
   */
  public function preRenderBuild(array $element) {
    $build = $element['#build'];
    $config = $build['config'];
    $config['current_path'] = Url::fromRoute('<current>')->toString();
    $goodies = $this->getSetting('goodies');

    $element['#config'] = $config;
    $element['#delta'] = $config['menu_name'];
    $element['#items'] = $this->buildMenuTree($config);
    $element['#attached']['library'][] = 'ultimenu/ultimenu';

    if (!empty($config['skin_basename'])) {
      $element['#attached']['library'][] = 'ultimenu/skin.' . $config['skin_basename'];
    }
    if (!empty($config['orientation']) && strpos($config['orientation'], 'v') !== FALSE) {
      $element['#attached']['library'][] = 'ultimenu/vertical';
    }
    if (!empty($config['ajaxify'])) {
      $element['#attached']['library'][] = 'ultimenu/ajax';
    }
    if (empty($goodies['no-extras'])) {
      $element['#attached']['library'][] = 'ultimenu/extras';
    }

    unset($build, $element['#build']);
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function buildMenuTree(array $config) {
    $menu_name = $config['menu_name'];
    $active_trails = $this->tree->getMenuActiveTrail()->getActiveTrailIds($menu_name);
    $ultimenu = [];

    foreach ($this->tree->loadMenuTree($menu_name) as $data) {
      // Generally we only deal with visible links, but just in case.
      if (!$data->link->isEnabled()) {
        continue;
      }

      if ($data->access !== NULL && !$data->access instanceof AccessResultInterface) {
        throw new \DomainException('MenuLinkTreeElement::access must be either NULL or an AccessResultInterface object.');
      }

      // Only render accessible links.
      if ($data->access instanceof AccessResultInterface && !$data->access->isAllowed()) {
        continue;
      }

      $ultimenu[$data->link->getPluginId()] = $this->buildMenuItem($data, $active_trails, $config);
    }
    return $ultimenu;
  }

  /**
   * {@inheritdoc}
   */
  public function buildMenuItem($data, array $active_trails, array $config) {
    $menu_name    = $config['menu_name'];
    $goodies      = $this->getSetting('goodies');
    $link         = $data->link;
    $key          = $link->getPluginId();
    $url          = $link->getUrlObject();
    $title        = Html::escape($link->getTitle());
    $has_children = $data->hasChildren;
    $expanded     = $link->isExpanded();
    $plugin_id    = $link->getPluginId();
    $show_submenu = !empty($config['submenu']) && $expanded && $has_children;
    $definition   = $link->getPluginDefinition();
    $route_name   = $definition['route_name'];
    $li_classes   = $li_attributes = [];
    $flyout       = '';

    if ($url->isExternal()) {
      $li_classes[] = 'external';
    }

    // @todo check provider: standard, menu_link_content.
    $shortened_hash = $this->tool->getShortenedHash($definition['id']);
    if ($definition['route_name'] == '<front>') {
      $shortened_uuid = 'front_page';
    }
    else {
      $shortened_uuid = $this->tool->getShortenedUuid($definition['id']);
    }

    // Titles.
    // Get the non-localized title to get relevant info.
    $title_html = $title = trim(strip_tags($title));

    // Allows simple icon addition via a piped title, e.g.: icon-home|Home.
    $icon = '';
    $titles = $this->extractTitle($title);
    if (!empty($titles['icon'])) {
      $title = $title_html = $titles['title'];
      $icon_class = $titles['fa'] ? 'fa ' . $titles['icon'] : 'icon ' . $titles['icon'];
      $icon = '<span class="ultimenu__icon ' . $icon_class . '" aria-hidden="true"></span>';
    }

    if (!empty($goodies['menu-desc']) && $description = $link->getDescription()) {
      // Render description, if so configured.
      // If you override this, be sure to have proper sanitization.
      $menu_description = '<small>' . strip_tags($description, '<em><strong><i><b>') . '</small>';
      $is_desc_top = isset($goodies['desc-top']) && $goodies['desc-top'];
      $title_html = $is_desc_top ? $menu_description . $title : $title . $menu_description;
    }

    // Holds the title in a separate SPAN for easy positioning if it has icon.
    if ($icon) {
      $title_html = $icon . '<span class="ultimenu__title">' . $title_html . '</span>';
    }

    // Must run after the title, modified, or not, the region depends on it.
    $children = $show_submenu ? $this->tree->loadSubMenuTree($menu_name, $plugin_id, $title) : [];
    $region = $this->tool->getRegionKey($menu_name, $plugin_id, $title, static::MAX_LENGTH);
    $config['ajax_cid'] = $menu_name . '|' . $plugin_id . '|' . $title;
    $config['ajax_rid'] = $region;
    $config['ajax_sub'] = $show_submenu ? 1 : 0;
    $config['is_ajax_region'] = FALSE;
    $config['is_active'] = array_key_exists($key, $active_trails);
    $config['route_name'] = $route_name;
    $config['uuid'] = $definition['id'];

    $link_options = $link->getOptions();
    if ($url->isRouted()) {
      if (array_key_exists($key, $active_trails)) {
        $li_classes[] = 'is-active-trail';
      }

      // Front page has no active trail.
      if ($route_name == '<front>') {
        // Intentionally on the second line to not hit it till required.
        if ($this->tool->getPathMatcher()->isFrontPage()) {
          $li_classes[] = 'is-active-trail';
        }
      }

      // Also enable set_active_class for the contained link.
      $link_options['set_active_class'] = TRUE;

      // Add a "data-drupal-link-system-path" attribute to let the
      // drupal.active-link library know the path in a standardized manner.
      $system_path = $url->getInternalPath();
      $system_path = $system_path == '' ? '<front>' : $system_path;

      // @todo System path is deprecated - use the route name and parameters.
      // Special case for the front page.
      $link_options['attributes']['data-drupal-link-system-path'] = $system_path;
      $config['system_path'] = $system_path;
    }

    // Remove browser tooltip if so configured.
    if (isset($goodies['no-tooltip']) && $goodies['no-tooltip']) {
      $link_options['attributes']['title'] = '';
    }

    // Add LI title class based on title if so configured.
    if (isset($goodies['title-class']) && $goodies['title-class']) {
      $li_classes[] = Html::cleanCssIdentifier(mb_strtolower('uitem--' . $title));
    }

    // Add LI counter class based on counter if so configured.
    if (isset($goodies['counter-class']) && $goodies['counter-class']) {
      static $item_id = 0;
      $li_classes[] = 'uitem--' . (++$item_id);
    }

    // Add LI mlid/shortened uuid class based on uuid if so configured.
    if (!empty($goodies['mlid-class']) && $shortened_uuid) {
      $li_classes[] = Html::cleanCssIdentifier(mb_strtolower('uitem--' . $shortened_uuid));
    }

    // Add LI mlid/shortened hash class based on hash if so configured.
    if (!empty($goodies['mlid-hash-class']) && $shortened_hash) {
      $li_classes[] = Html::cleanCssIdentifier(mb_strtolower('uitem--' . $shortened_hash));
    }

    // Handle list item attributes.
    $li_attributes['class'] = array_merge(['ultimenu__item', 'uitem'], $li_classes);

    // Flyout.
    if ($regions = $this->getSetting('regions')) {
      if (!empty($regions[$region])) {

        // Provides fallback with a link in case AJAX fails.
        if (empty($config['ajaxify'])) {
          $flyout = $this->buildFlyout($region, $config, $children);
        }
        else {
          // We have a mix of (non-)ajaxified blocks here.
          $ajax_regions = isset($config['regions']) ? array_filter($config['regions']) : [];
          $config['is_ajax_region'] = $regions && in_array($region, $ajax_regions);
          $flyout = $config['is_ajax_region'] ? $this->buildAjaxLink($config) : $this->buildFlyout($region, $config, $children);
        }
      }
    }

    // Provides hints for AJAX.
    $flyout_attributes = [];
    if (!empty($flyout)) {
      if ($config['is_ajax_region'] && $config['ajax_rid']) {
        $flyout_attributes['data-ultiajax-region'] = $config['ajax_rid'];
        $link_options['attributes']['data-ultiajax-trigger'] = TRUE;
      }
      $title_html .= '<span class="caret" area-hidden="true"></span>';
    }

    $extra_classes = isset($link_options['attributes']['class']) ? $link_options['attributes']['class'] : [];
    $link_options['attributes']['class'] = $extra_classes ? array_merge(['ultimenu__link'], $extra_classes) : ['ultimenu__link'];

    $link_element = [
      '#type' => 'link',
      '#options' => $link_options,
      '#url' => $url,
      '#title' => [
        '#markup' => $title_html,
        '#allowed_tags' => ['b', 'em', 'i', 'small', 'span', 'strong'],
      ],
    ];

    // Pass link to template.
    return [
      'link' => $link_element,
      'flyout' => $flyout,
      'attributes' => new Attribute($li_attributes),
      'flyout_attributes' => new Attribute($flyout_attributes),
      'config' => $config,
    ];
  }

  /**
   * Returns title with an icon class if available, e.g.: fa-mail|Contact us.
   */
  public function extractTitle($title) {
    $is_icon = substr($title, 0, 5) === 'icon-';
    $is_fontawesome = substr($title, 0, 3) === 'fa-';
    if (strpos($title, '|') !== FALSE && ($is_icon || $is_fontawesome)) {
      list($icon_class, $title) = array_pad(array_map('trim', explode("|", $title, 2)), 2, NULL);
      return ['icon' => $icon_class, 'title' => $title, 'fa' => $is_fontawesome];
    }
    return ['title' => $title];
  }

  /**
   * {@inheritdoc}
   */
  public function buildAjaxLink(array $config = []) {
    return [
      '#type' => 'link',
      '#title' => $this->getFallbackText(),
      '#attributes' => [
        'class' => [
          'ultiajax__fallback',
          'use-ajax',
        ],
        'rel' => 'nofollow',
      ],
      '#url' => $this->getAjaxUrl($config),
    ];
  }

  /**
   * Returns the AJAX URL.
   */
  public function getAjaxUrl(array $config = []) {
    return Url::fromRoute('ultimenu.ajax', [
      'cid' => $config['ajax_cid'],
      'cur' => $config['current_path'],
      'sub' => $config['ajax_sub'],
    ]);
  }

  /**
   * Return the fallback text.
   */
  public function getFallbackText() {
    return $this->t('@fallback_text', ['@fallback_text' => $this->getSetting('fallback_text') ?: 'Loading... Click here if it takes longer.']);
  }

  /**
   * {@inheritdoc}
   */
  public function buildFlyout($region, array $config, array $children = []) {
    $build   = $blocks = $content = [];
    $reverse = FALSE;

    if ($children) {
      $reverse = !empty($config['submenu_position']) && $config['submenu_position'] == 'bottom';
      $content[] = $children;
    }

    if ($blocks = $this->getBlocksByRegion($region, $config)) {
      $content[] = $blocks;
    }

    if ($content = array_filter($content)) {
      $build['content'] = $reverse ? array_reverse($content, TRUE) : $content;
      $build['#config'] = $config;
      $build['#region'] = $region;
      $build['#sorted'] = TRUE;

      // Add the region theme wrapper for the Ultimenu flyout.
      $build['#theme_wrappers'][] = 'region';
    }
    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function getBlocksByRegion($region, array $config) {
    if (!isset($this->blocks[$region])) {
      $build = [];
      $theme = $this->getConfig('system.theme')->get('default');
      $blocks = $this->entityTypeManager->getStorage('block')->loadByProperties([
        'theme' => $theme,
        'region' => $region,
      ]);

      if ($blocks) {
        uasort($blocks, 'Drupal\block\Entity\Block::sort');

        // Only provides extra access checks if using ajaxified blocks.
        if (empty($config['ajaxify'])) {
          foreach ($blocks as $key => $block) {
            if ($block->access('view')) {
              $build[$key] = $this->entityTypeManager->getViewBuilder($block->getEntityTypeId())->view($block, 'block');
            }
          }
        }
        else {
          foreach ($blocks as $key => $block) {
            if ($this->tool->isAllowedBlock($block, $config)) {
              $build[$key] = $this->entityTypeManager->getViewBuilder($block->getEntityTypeId())->view($block, 'block');
            }
          }
        }
      }
      $this->blocks[$region] = $build;
    }
    return $this->blocks[$region];
  }

  /**
   * {@inheritdoc}
   */
  public function getRegions() {
    if (!isset($this->regions)) {
      $blocks      = $this->getSetting('blocks');
      $menu_blocks = is_array($blocks) ? array_filter($blocks) : [$blocks];
      $menus       = [];

      foreach ($menu_blocks as $delta => $title) {
        $menus[$delta] = $this->tree->loadMenuTree($delta);
      }

      // Allow alteration.
      $this->moduleHandler->alter('ultimenu_menus_info', $menus);

      $this->regions = [];
      foreach ($menus as $menu_name => $tree) {
        $tree = $this->tree->getMenuTree()->build($tree);
        if (isset($tree['#items'])) {
          foreach ($tree['#items'] as $key => $item) {
            $name_id = $this->tool->truncateRegionKey($menu_name);
            $name_id_nice = str_replace("_", " ", $name_id);

            if ($item['title']) {
              $menu_title = strip_tags($item['title']);
              $titles = $this->extractTitle($menu_title);
              $menu_title = $titles['title'];
              $region_key = $this->tool->getRegionKey($menu_name, $key, $menu_title, static::MAX_LENGTH);
              $this->regions[$region_key] = "Ultimenu:$name_id_nice: $menu_title";
              $this->regions = array_unique($this->regions);
            }
          }
        }
      }

      // Allow alteration.
      $this->moduleHandler->alter('ultimenu_regions_info', $this->regions);
    }
    return $this->regions;
  }

  /**
   * {@inheritdoc}
   */
  public function getEnabledRegions() {
    if (!isset($this->enabledRegions)) {
      $this->enabledRegions = [];
      $regions_all = $this->getRegions();

      // First limit to enabled regions from the settings.
      if (($regions_enabled = $this->getSetting('regions')) !== NULL) {
        foreach (array_filter($regions_enabled) as $enabled) {
          // We must depends on enabled menu items as always.
          // A disabled menu item will automatically drop its Ultimenu region.
          if (array_key_exists($enabled, $regions_all)) {
            $this->enabledRegions[$enabled] = $regions_all[$enabled];
          }
        }
      }
    }
    return $this->enabledRegions;
  }

  /**
   * {@inheritdoc}
   */
  public function removeRegions() {
    $goodies = $this->getSetting('goodies');
    if (!empty($goodies['force-remove-region']) && $theme_regions = $this->tool->parseThemeInfo($this->getRegions())) {
      return $theme_regions;
    }
    return FALSE;
  }

}
